// Copyright (c) Microsoft Open Technologies, Inc. All rights reserved.
// Licensed under the Apache License, Version 2.0. See License.txt in the project root for license information.

using System.Threading;
using System.Threading.Tasks;
using Microsoft.Data.Entity.Metadata;
using Microsoft.Data.Entity.Utilities;

namespace Microsoft.Data.Entity.Relational
{
    public abstract class RelationalDataStoreCreator : IRelationalDataStoreCreator
    {
        public abstract bool Exists();

        public abstract Task<bool> ExistsAsync(CancellationToken cancellationToken = default(CancellationToken));

        public abstract void Create();

        public abstract Task CreateAsync(CancellationToken cancellationToken = default(CancellationToken));

        public abstract void Delete();

        public abstract Task DeleteAsync(CancellationToken cancellationToken = default(CancellationToken));

        public abstract void CreateTables(IModel model);

        public abstract Task CreateTablesAsync(IModel model, CancellationToken cancellationToken = default(CancellationToken));

        public abstract bool HasTables();

        public abstract Task<bool> HasTablesAsync(CancellationToken cancellationToken = default(CancellationToken));

        public virtual bool EnsureDeleted(IModel model)
        {
            Check.NotNull(model, nameof(model));

            if (Exists())
            {
                Delete();
                return true;
            }
            return false;
        }

        public virtual async Task<bool> EnsureDeletedAsync(IModel model, CancellationToken cancellationToken = default(CancellationToken))
        {
            Check.NotNull(model, nameof(model));

            if (await ExistsAsync(cancellationToken).WithCurrentCulture())
            {
                await DeleteAsync(cancellationToken).WithCurrentCulture();

                return true;
            }
            return false;
        }

        public virtual bool EnsureCreated(IModel model)
        {
            Check.NotNull(model, nameof(model));

            if (!Exists())
            {
                Create();
                CreateTables(model);
                return true;
            }

            if (!HasTables())
            {
                CreateTables(model);
                return true;
            }

            return false;
        }

        public virtual async Task<bool> EnsureCreatedAsync(IModel model, CancellationToken cancellationToken = default(CancellationToken))
        {
            Check.NotNull(model, nameof(model));

            if (!await ExistsAsync(cancellationToken).WithCurrentCulture())
            {
                await CreateAsync(cancellationToken).WithCurrentCulture();
                await CreateTablesAsync(model, cancellationToken).WithCurrentCulture();

                return true;
            }

            if (!await HasTablesAsync(cancellationToken).WithCurrentCulture())
            {
                await CreateTablesAsync(model, cancellationToken).WithCurrentCulture();

                return true;
            }

            return false;
        }
    }
}
